# -*- coding: utf-8 -*-

"""IDLE extension showing source code as a interactive tree.

To install and enable this extension,
- find the directory named idlelib in your Python installation,
- drop this file you are reading in the directory you found .../idlelib,
- add the text below in .../idlelib/config-extensions.def.

[ClickableTree]
enable=1
enable_shell=0
language=fr
[ClickableTree_cfgBindings]
show-tree=<Alt-Key-k>

See below the TRANSLATION_DICT for available languages.

Then start IDLE, open a new window (editor). You should see some Python code.
Pressing Alt-K or clicking the Edit->Clickable Tree shows a tree that you
can modify (click a string, click it again, edit and confirm with Return.

TODO:
- PythonBox -> filename
- no tree update if .:; (<Any-KeyRelease> not even reacting)
- code templates src -> view, view -> src
- alt K focuses on the menus under Windows
- tree is expanded fully at each source modification
- parse/unparse Python code with neat tools
  codegen: https://gist.github.com/791312
  ast pretty printer: http://code.activestate.com/recipes/533145-ast-pretty-printer/
  simplest ast visitor: http://snippets.dzone.com/posts/show/3360
- improve the UI (left-click, right-click, popups,
  keybindings to hide, navigate, delete insert, copy, cut, paste,
  dropdown with vars in scope...)
- improve the 'text modified' detector
  http://code.activestate.com/recipes/464635-call-a-callback-when-a-tkintertext-is-modified/
- fix the OO design (passing refs around was easy but spaghetti)
- syntax highlighting in the tree
- show exceptions debug info, breakpoints, markers in the tree
- do something meaningful if user wants the tree and tree already shown

Thanks to:
- http://code.activestate.com/recipes/259116-tkinter-tree-widget-using-idlelibtreewidget/
"""

import sys
import time
P2 = sys.version_info < (3, 0)
if P2:
    import Tkinter as Tk
    from PyShell import use_subprocess
else:
    import tkinter as Tk
    from .PyShell import use_subprocess
from idlelib.TreeWidget import TreeItem, TreeNode
from idlelib.configHandler import idleConf

def dbg(txt):
    "May print 'txt' to the console. Used for dev only."
    global use_subprocess
    if 1 and use_subprocess:
        print(txt)

TEST_DATA = dict()
TEST_DATA['Python'] = '''def hello(name):
    if len(name):
        print 'hello ' + name
    else:
        print 'Hey! Your name!?'
'''
PYTHON_INDENT = "    "

TEST_MODE = 'Python_Lines'
TEST_LANG = '_' not in TEST_MODE and \
                TEST_MODE or TEST_MODE[:TEST_MODE.index('_')]

UI_LANGUAGE = idleConf.GetOption('extensions','ClickableTree',
           'language', default='en')

TRANSLATION_DICT = {
        'fr' : {
            "File" : "Fichier",
            "Clic_kable Tree" : "Arbre Clic_kable",
            "Hide Tree" : "Cacher l'arbre",
            "Tree already shown." : "Arbre déjà en vue.",
            "Delete" : "Supprimer",
            "Insert" : "Insérer",
            "Insert above" : "Insérer au-dessus",
            "Insert below" : "Insérer en-dessous",
            "insert your code here" : "insérez votre code ici",
            "Cancel" : "Annuler",
            }
        }

def _(txt):
    """gettext-like translation function."""

    if UI_LANGUAGE == 'en':
        return txt
    try:
        transdict = TRANSLATION_DICT[UI_LANGUAGE]
    except KeyError:
        print("Language not supported: %s." % UI_LANGUAGE)
        return txt
    try:
        return transdict[txt]
    except KeyError:
        print("Not translated in %s: '%s'." % (UI_LANGUAGE, txt))
        return txt


class ClickableTree():
    "IDLE extension class for work with an interactive source tree."

    # Insert a menu item.
    menudefs = [
        ('edit', [
            (_("Clic_kable Tree"), '<<show-tree>>'),
            ])
        ]

    def __init__(self, editwin):
        self.editwin = editwin
        # self.editwin.text.insert('1.0', TEST_DATA[TEST_LANG])
        self.treewindow = TreeWindow(self.editwin)

    def show_tree_event(self, event):
        if self.treewindow.hidden:
            self.treewindow.showTree()
        else:
            dbg(_("Tree already shown."))


class AbstractTreeItem(TreeItem):

    def __init__(self, docstruct):
        self.docstruct = docstruct
        self.rootnode = None

    def GetText(self):
        # Please implement in a subclass.
        pass

    def SetText(self):
        # Please implement in a subclass.
        pass

    def IsExpandable(self):
        return self.docstruct.hasChildNodes()

    def IsEditable(self):
        return True

    def GetSubList(self):
        children = self.docstruct.childNodes
        item_class = eval(TEST_LANG + 'TreeItem')
        prelist = [item_class(dsnode) for dsnode in children]
        # Give all children the rootnode ref to allow notifications.
        for dsnode in prelist:
            dsnode.rootnode = self.rootnode
        return [dsnode for dsnode in prelist if dsnode.GetText().strip()]


class ParseSourceError(Exception):

    def __init__(self, msg, line, col):
        self.msg = msg
        self.line = line
        self.col = col


class PythonTreeItem(AbstractTreeItem):

    def GetText(self):
        return self.docstruct.text

    def SetText(self, txt):
        self.docstruct.text = txt
        self.rootnode.itemChanged()


class AbstractSourceNode(TreeNode):

    def __init__(self, canvas, parent, item):
        TreeNode.__init__(self, canvas, parent, item)
        self.image_id = 0  # hack to avoid drawing icons

    def itemChanged(self):
        rootnode = self.item.rootnode
        content = rootnode.getTreeSource()
        treewin = self.item.rootnode.treewin
        treewin.editwin.text.delete('1.0', 'end')
        treewin.editwin.text.insert('1.0', content)
        treewin.editwin_cache = content

    def updateAll(self):
        # Order matters since the tree is built
        # parsing the source in the edit window.
        # First: update the paired edit window.
        self.item.rootnode.treewin.editwin.text.undo_block_start()
        self.itemChanged()
        self.item.rootnode.treewin.editwin.text.undo_block_stop()
        # Then update the tree.
        self.item.rootnode.treewin.editwin_cache = ''
        self.item.rootnode.treewin.update_editwin()

    def getTreeSource(self):
        # Please implement in a subclass.
        pass

    def contextmenu(self, event=None):
        self.menu = Tk.Menu(self.canvas)
        prefix = 'context__'
        text_offset = 3  # sorting numbers are 2 chars long
        commands_as_strings = [attr for attr in dir(self) if attr.startswith(prefix)]
        submenus = dict()
        for command_string in commands_as_strings:
            command = getattr(self, command_string)
            command_clean_string = command_string[len(prefix) + text_offset:]
            if "__" not in command_clean_string:
                label = _(command_clean_string.replace("_", " "))
                self.menu.add_command(label=label, command=command)
            else:
                submenu_as_string, __, raw_label = command_clean_string.partition('__')
                label = _(raw_label[text_offset:].replace("_", " "))
                if not submenus.has_key(submenu_as_string):
                    submenus[submenu_as_string] = Tk.Menu(self.menu)
                    self.menu.add_cascade(menu=submenus[submenu_as_string], label=_(submenu_as_string))
                    submenus[submenu_as_string].add_command(label=_("Cancel"), command=None)
                submenus[submenu_as_string].add_command(label=label, command=command)
        self.menu.post(event.x_root, event.y_root)

    def context__01_Cancel(self):
        """Cancel the context menu operation."""
        pass

    def context__02_Delete(self):
        """Delete the node from the context menu."""
        self.parent.item.docstruct.childNodes.remove(self.item.docstruct)
        self.updateAll()

    def context__03_Insert__01_Insert_above(self):
        """Insert a node above the selected from the context menu."""
        siblings = self.parent.item.docstruct.childNodes
        current_idx = siblings.index(self.item.docstruct)
        siblings.insert(current_idx, Python_LinesDocStruct(
           source = '',  # no child for this new node
           level = self.item.docstruct.level,
           text = '# ' + _("insert your code here")))
        self.updateAll()

    def context__03_Insert__02_Insert_below(self):
        """Insert a node below the selected from the context menu."""
        siblings = self.parent.item.docstruct.childNodes
        current_idx = siblings.index(self.item.docstruct)
        siblings.insert(current_idx + 1, Python_LinesDocStruct(
           source = '',  # no child for this new node
           level = self.item.docstruct.level,
           text = '# ' + _("insert your code here")))
        self.updateAll()

    def select(self, event=None):
        if self.selected:
            return
        self.deselectall()
        self.selected = True
        self.drawtext()

    def select_and_menu(self, event=None):
        self.select()
        # Wait for drawtext to finish in self.select.
        self.canvas.after(100, self.contextmenu, event)

    def expandRecursively(self):
        """Recursively expand all nodes, self and children, and...

        children of children..."""
        self.expand()
        for child in self.children:
            child.expandRecursively()

    def draw(self, x, y):
        # XXX This hard-codes too many geometry constants!
        self.x, self.y = x, y
        self.drawtext()
        if self.state != 'expanded':
            return y+17
        # draw children
        if not self.children:
            sublist = self.item._GetSubList()
            if not sublist:
                # _IsExpandable() was mistaken; that's allowed
                return y+17
            for item in sublist:
                child = self.__class__(self.canvas, self, item)
                self.children.append(child)
        cx = x+20  # horizontal child offset
        cy = y+17  # vertical child offset
        cylast = 0
        for child in self.children:
            cylast = cy
            self.canvas.create_line(x+9, cy+7, cx, cy+7, fill="gray50")
            cy = child.draw(cx, cy)
            if child.item._IsExpandable():
                if child.state == 'expanded':
                    iconname = "minusnode"
                    callback = child.collapse
                else:
                    iconname = "plusnode"
                    callback = child.expand
                image = self.geticonimage(iconname)
                id = self.canvas.create_image(x+9, cylast+7, image=image)
                # XXX This leaks bindings until canvas is deleted:
                self.canvas.tag_bind(id, "<1>", callback)
        id = self.canvas.create_line(x+9, y+10, x+9, cylast+7,
            ##stipple="gray50",     # XXX Seems broken in Tk 8.0.x
            fill="gray50")
        self.canvas.tag_lower(id) # XXX .lower(id) before Python 1.5.2
        return cy

    def drawtext(self):
        textx = self.x+1
        texty = self.y-1
        labeltext = self.item.GetLabelText()
        if labeltext:
            id = self.canvas.create_text(textx, texty, anchor="nw",
                                         text=labeltext)
            x0, y0, x1, y1 = self.canvas.bbox(id)
            textx = max(x1, 200) + 10
        text = self.item.GetText() or "<no text>"
        try:
            self.entry
        except AttributeError:
            pass
        else:
            self.edit_finish()
        try:
            label = self.label
        except AttributeError:
            # padding carefully selected (on Windows) to match Entry widget:
            self.label = Tk.Label(self.canvas, text=text, bd=0, padx=2, pady=2)
        theme = idleConf.GetOption('main','Theme','name')
        if self.selected:
            self.label.configure(idleConf.GetHighlight(theme, 'hilite'))
        else:
            self.label.configure(idleConf.GetHighlight(theme, 'normal'))
        self.label.configure(font=("Courier", 10))
        id = self.canvas.create_window(textx, texty,
                                       anchor="nw", window=self.label)
        self.label.bind("<Button-1>", self.select_or_edit)
        if (self.canvas.tk.call('tk', 'windowingsystem')=='aqua'):
            self.label.bind('<Button-2>', self.select_and_menu)
            self.label.bind('<Control-1>', self.select_and_menu)
        else:
            self.label.bind('<Button-3>', self.select_and_menu)
        self.text_id = id

    def edit(self, event=None):
        self.entry = Tk.Entry(self.label, bd=0, highlightthickness=1, width=0)
        self.entry.configure(font=("Courier", 10))
        self.entry.insert(0, self.label['text'])
        self.entry.selection_range(0, Tk.END)
        self.entry.pack(ipadx=5)
        self.entry.focus_set()
        self.entry.bind("<Return>", self.edit_finish)
        self.entry.bind("<Escape>", self.edit_cancel)


class PythonSourceNode(AbstractSourceNode):

    def getTreeSource(self):
        return self.item.docstruct.topython()


class Python_LinesDocStruct(object):
    """DocStruct that dumbly parses Python code line by line.

    TODO: merge backslashes.
    Originally built to mimic a dom.documentElement , where dom is returned by
    xml.dom.minidom.parseString."""

    def __init__(self, source, level = 0, text = 'PythonBox'):
        self.level = level
        self.text = text
        self.childNodes = []
        lines = source.split('\n')
        # lines beginning with no whitespace are the text of the children
        children_line_nums = [line for line in range(len(lines)) \
                                   if len(lines[line]) and lines[line][0] not in [' ', '\t']]
        for line_num in children_line_nums:
            start = line_num + 1
            if line_num < children_line_nums[-1]:
                # between two children
                end = children_line_nums[children_line_nums.index(line_num) + 1]
            else:
                # last child, until the end of soure
                end = len(lines)
            # dedent lines for the next level
            new_lines = [line[4:] for line in lines[start:end]]
            self.childNodes.append(
               Python_LinesDocStruct(
                  '\n'.join(new_lines),
                  self.level + 1,
                  lines[line_num]))

    def topython(self):
        ret = ""
        if self.level:  # The root elt (level 0) has no relevant text.
            indent = (self.level - 1) * PYTHON_INDENT
            ret = ret + indent + self.text + '\n'
        for child in self.childNodes:
            ret = ret + child.topython() + '\n'
        return ret[:-1]  # Remove \n from each sub tree.

    def hasChildNodes(self):
        return len(self.childNodes)


class TreeWindow():

    def __init__(self, editwin):
        self.editwin = editwin
        self.editwin_cache = ""
        self.update_rate = 1  # in seconds
        self.latest_update_request_time = 0
        self.latest_update_request_id = None
        self.hidden = True

    def showTree(self):
        self.win = Tk.Tk()
        self.win.option_add('*tearOff', False)
        menubar = Tk.Menu(self.win)
        self.win['menu'] = menubar
        self.win.protocol('WM_DELETE_WINDOW', self.hideTree)
        menu_file = Tk.Menu(menubar)
        menu_file.add_command(label=_("Hide Tree"), command=self.hideTree)
        menubar.add_cascade(menu=menu_file, label=_("File"))
        self.canvas = Tk.Canvas(self.win, bg="white", highlightthickness=0,
                                          width=800, height=600)
        #self.canvas.grid(row=0, column=0, sticky="nsew")
        self.canvas.pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)
        # Scroll bars creation and gridation.
        self.vbar = Tk.Scrollbar(self.win, name="vbar")
        #self.vbar.grid(row=0, column=1, sticky="nse")
        self.vbar.pack(side=Tk.RIGHT, fill=Tk.Y)
        self.hbar = Tk.Scrollbar(self.win, name="hbar", orient="horizontal")
        #self.hbar.grid(row=1, column=0, sticky="ews")
        self.hbar.pack(side=Tk.BOTTOM, fill=Tk.X)
        # Scroll bar relations with the canvas.
        self.canvas['yscrollcommand'] = self.vbar.set
        self.vbar['command'] = self.canvas.yview
        self.canvas['xscrollcommand'] = self.hbar.set
        self.hbar['command'] = self.canvas.xview
        # Last things.
        self.update_editwin()
        self.startTracking()
        self.hidden = False

    def hideTree(self):
        self.stopTracking()
        self.editwin_cache = ""
        self.win.destroy()
        self.hidden = True

    def update_requested(self, event):
        dbg("update requested")
        self.latest_update_request_time = time.time()
        if self.latest_update_request_id:
            self.win.after_cancel(self.latest_update_request_id)
        self.latest_update_request_id = \
           self.win.after(self.update_rate * 1000, self.update_editwin)

    def update_editwin(self):
        current_win_content = self.getEditWinSource()
        if current_win_content.strip() == self.editwin_cache.strip():
            # Tree still in sync, no need to redraw it.
            return
        self.editwin_cache = current_win_content
        self.clear_errors()
        item = None
        item_class = eval(TEST_LANG + 'TreeItem')
        docstruct_class = eval(TEST_MODE + 'DocStruct')
        try:
            item = item_class(docstruct_class(current_win_content))
        except ParseSourceError as e:
            self.error(e.msg, e.line, e.col)
        if item:
            sourcenode_class = eval(TEST_LANG + 'SourceNode')
            self.rootnode = sourcenode_class(self.canvas, None, item)
            # Give references to allow notifications.
            item.rootnode = self.rootnode
            self.rootnode.treewin = self
            # Prepare for first display.
            self.rootnode.update()
            self.rootnode.expandRecursively()

    def startTracking(self):
        self.editwin.text.bind('<Any-KeyRelease>', self.update_requested)
        self.editwin_cache = self.getEditWinSource()

    def stopTracking(self):
        self.editwin.text.bind('<Any-KeyRelease>', lambda x: None)

    def getEditWinSource(self):
        return self.editwin.text.get('1.0', 'end')

    def error(self, error_msg, line, column):
        "Show an error in the paired edit window."
        start = line + '.' + column
        end = line + '.end'
        self.editwin.text.tag_add('error', start, end)
        self.editwin.text.tag_configure('error', background='yellow')
        # TODO: find a way to show the error msg.
        dbg("An error occured:" + error_msg)

    def clear_errors(self):
        "Remove any error marker in the paired edit window."
        self.editwin.text.tag_delete('error')
