# -*- coding: utf-8 -*-

"""IDLE extension to interpret home-made languages via Python.

To install and enable this extension,
- find the directory named idlelib in your Python installation,
- drop this file you are reading in the directory you found .../idlelib,
- drop calto.py in the same directory,
- add the text below in .../idlelib/config-extensions.def.

[Tirtal]
enable=1
enable_shell=0
[Tirtal_cfgBindings]
show-screen=<Alt-Key-b>

Start IDLE, open a new window (editor). You should see some calculator code.
Pressing Alt-B or clicking the Edit->Tirtal shows a fake calculator screen.

"""

import sys
import re
P2 = sys.version_info < (3, 0)
if P2:
    import Tkinter as Tk
    from PyShell import use_subprocess
    import calto
else:
    import tkinter as Tk
    from .PyShell import use_subprocess
    from . import calto
from idlelib.configHandler import idleConf

def dbg(txt):
    "May print 'txt' to the console. Used for dev only."
    global use_subprocess
    if 1 and use_subprocess:
        print(txt)


UI_LANGUAGE = idleConf.GetOption('extensions','ClickableTree',
           'language', default='en')

TRANSLATION_DICT = {
        'fr' : {
            "File" : "Fichier",
            "Tir_tal" : "Tir_tal",
            "Hide Screen" : "Fermer l’écran",
            "Run Program" : "Lancer le programme",
            "Cancel" : "Annuler",
            }
        }


def _(txt):
    """gettext-like translation function."""

    if UI_LANGUAGE == 'en':
        return txt
    try:
        transdict = TRANSLATION_DICT[UI_LANGUAGE]
    except KeyError:
        print("Language not supported: %s." % UI_LANGUAGE)
        return txt
    try:
        return transdict[txt]
    except KeyError:
        print("Not translated in %s: '%s'." % (UI_LANGUAGE, txt))
        return txt


class Tirtal():
    "IDLE extension class for work with home-made languages."

    # Insert a menu item.
    menudefs = [
        ('edit', [
            (_("Tir_tal"), '<<show-screen>>'),
            ])
        ]

    def __init__(self, editwin):
        self.editwin = editwin
        self.screenwindow = ScreenWindow(self.editwin)

    def show_screen_event(self, event):
        if self.screenwindow.hidden:
            self.screenwindow.showScreen()
        else:
            dbg(_("Screen already shown."))


class ParseSourceError(Exception):

    def __init__(self, msg, line, col):
        self.msg = msg
        self.line = line
        self.col = col


class ScreenWindow():

    def __init__(self, editwin):
        self.editwin = editwin
        self.hidden = True

    def showScreen(self):
        self.win = Tk.Tk()
        self.win.option_add('*tearOff', False)
        menubar = Tk.Menu(self.win)
        self.win['menu'] = menubar
        self.win.protocol('WM_DELETE_WINDOW', self.hideScreen)
        menu_file = Tk.Menu(menubar)
        menu_file.add_command(label=_("Hide Screen"), command=self.hideScreen)
        menu_file.add_command(label=_("Run Program"), command=self.runProgram)
        menubar.add_cascade(menu=menu_file, label=_("File"))
        self.text = Tk.Text(self.win, bg="white", highlightthickness=0,
                                          width=80, height=40,
                                          state='disabled')
        self.text.grid(row=0, column=0, sticky="nsew")
        # Scroll bars creation and gridation.
        self.vbar = Tk.Scrollbar(self.win, name="vbar")
        self.vbar.grid(row=0, column=1, sticky="nse")
        self.hbar = Tk.Scrollbar(self.win, name="hbar", orient="horizontal")
        self.hbar.grid(row=1, column=0, sticky="ews")
        # Scroll bar relations with the text widget.
        self.text['yscrollcommand'] = self.vbar.set
        self.vbar['command'] = self.text.yview
        self.text['xscrollcommand'] = self.hbar.set
        self.hbar['command'] = self.text.xview
        # Last things.
        self.hidden = False

    def hideScreen(self):
        self.win.destroy()
        self.hidden = True

    def runProgram(self):
        my_calto = calto.Casio(self)
        my_calto.run(self.getEditWinSource())

    def getEditWinSource(self):
        return self.editwin.text.get('1.0', 'end').strip()

    def error(self, error_msg, line, column):
        "Show an error in the paired edit window."
        start = line + '.' + column
        end = line + '.end'
        self.editwin.text.tag_add('error', start, end)
        self.editwin.text.tag_configure('error', background='yellow')
        # TODO: find a way to show the error msg.
        dbg("An error occured:" + error_msg)

    def clear_errors(self):
        "Remove any error marker in the paired edit window."
        self.editwin.text.tag_delete('error')
