"""A parser for calculator programming languages.

The goal is to provide a friendly environment to work with the programming
languages of some calculators. Now you can write programs in a real text
editor, then store, sort, show, print and share them easily.

Run a program in myfile.txt with:
python calto.py myfile.txt

Run the tests with:
python calto.py

Please note that the syntax is slightly different from the original
languages. See below.

All calculators:
    The : separator cannot be used anywhere else (e.g. in strings, comments).
    Python-like comments are available.
    Free indentation is allowed and has no influence.
    Maths expressions are Python expressions, so be careful with:
        / in Python 2 which always returns integers

See the main docstring of the relevant class for information about
a precise Calculator.

"""

import sys
import os
import re
P2 = sys.version_info < (3, 0)
try:
    if P2:
        import Tkinter as Tk
        import tkSimpleDialog as simpledialog
    else:
        import tkinter as Tk
        import tkinter.simpledialog as simpledialog
except ImportError:
    print("Tk not installed.")

if P2:
    SPLIT_SEP = unicode('|')
    def next(iterator):
        return iterator.next()
else:
    SPLIT_SEP = '|'

DEBUG = False

def dbg(txt):
    "May print 'txt' to the console. Used for dev only."
    if DEBUG:
        print(txt)

def mysplit(tosplit, splitchars):
    # http://boredzo.org/blog/category/programming/python (Monday, June 16th, 2008)
    transtab = dict(zip(map(ord, splitchars), SPLIT_SEP * len(splitchars)))
    try:
        tosplit = unicode(tosplit)
    except NameError:
        pass
    return tosplit.translate(transtab).split(SPLIT_SEP)

def string_as_int_or_float(txt):
    if txt.isdigit():
        return int(txt)
    else:
        return float(txt)


class Calto(object):

    def __init__(self, screen = None):
        self.screen = screen
        # Syntax members.
        # Where to split source code to make lines.
        self.split_chars = '\n'
        # Regex pattern for label lines to match.
        self.label_pattern = ''
        # Elts are [pat_as_str, self.action, compiled_pat].
        self.parser_list = []
        # Start with vanilla state.
        self.reset()
        # Please note that each implementation has different loopers,
        # and need to reset them in their own __init__, after having
        # defined the loopers.

    def reset(self):
        self.testing = False
        self.vars = dict()
        self.current_line = 0
        self.labels = dict()
        self.elses_for_ifs = dict()
        self.ifends_for_elses = dict()
        self.play_this_else = None
        self.displays = []
        if hasattr(self, 'loopers'):
            self.reset_loopers()

    def reset_loopers(self):
        self.loopers_stacks = dict()
        self.loopers_pointers = dict()
        loopers_names = [looper_triple[0] for looper_triple in self.loopers]
        for looper in loopers_names:
            self.loopers_stacks.update({looper: []})
            self.loopers_pointers.update({looper: dict()})

    def run(self, source):
        self.lines = [line for line in mysplit(source, self.split_chars) \
                                       if line != '']
        self.parse_labels()
        self.parse_structures()
        # Parse and run the source code.
        while self.current_line < len(self.lines):
            line = self.lines[self.current_line].strip()
            dbg(' ' * 10 + line)
            # parse_line has desired side effects on the screen,
            # only self.display accumulates results (for tests).
            result = self.parse_line(line)
            # The line pointer may have changed, but may have been thrown
            # one line after the last (IfEnd at the end of file).
            if self.current_line < len(self.lines):
                line = self.lines[self.current_line]
            else:
                dbg("Ignore being thrown after the last one.")
            # We may force to display the last result:
            if self.current_line == len(self.lines) - 1 and \
               self.display_if_last_expr(line):
                self.display(result)
            self.current_line = self.current_line + 1

    def parse_labels(self):
        for line_num, line in enumerate(self.lines):
            comp_pattern = re.compile(self.label_pattern)
            m = comp_pattern.match(line)
            if m:
                try:
                    label_id = int(m.group(1))
                except Exception:
                    raise
                self.labels[label_id] = line_num

    def terminate(self):
        self.current_line = len(self.lines)

    def display_if_last_expr(self, line):
        end_instructions = [looper_triple[2] for looper_triple in self.loopers]
        for end_instruction in end_instructions:
            if line == end_instruction:
                return False
        return not any([
                line.startswith('#'),
                line[0] == '"' and line[-1] == '"',
                line.startswith(self.ifend_str),
                ])

    def parse_line(self, line):
        for parser_item in self.parser_list:
            comp_pattern, func = parser_item[2], parser_item[1]
            m = comp_pattern.match(line)
            if m:
                return func(m.groups())
        # No match so far, try to evaluate the line as an expression.
        return self.evaluate(line)

    def display(self, smthing):
        """Convert to str, append to self.displays and display to user."""
        text = str(smthing)
        self.displays.append(text)
        if DEBUG or not self.testing:
            if self.screen:
                self.screen.text.config(state='normal')
                self.screen.text.insert('end', text + '\n')
                self.screen.text.config(state='disabled')
            else:
                print(text)

    def prompt(self, params):
        txt, varname = params
        val = self.ask(txt)
        self.display(txt)
        self.vars[varname] = string_as_int_or_float(val)
        return self.vars[varname]

    def ask(self, question, default=""):
        user_input = ""
        while not user_input:
            if self.screen:
                user_input = simpledialog.askstring("Calto", question,
                                 parent=self.screen.win, initialvalue=default)
            else:
                user_input = raw_input(question)
        return user_input

    def compile_patterns(self):
        for parser_item in self.parser_list:
            parser_item.append(re.compile(parser_item[0]))

    def evaluate(self, expr):
        """Return a safe eval'd expr (maybe not a str), using self.vars."""
        # See http://code.activestate.com/recipes/496746-restricted-safe-eval/
        # Or simpler, from http://docs.python.org/library/functions.html#eval
        # And better, from http://lybniz2.sourceforge.net/safeeval.html (TODO)
        env = {'__builtins__': None, 'int': int}
        env.update(self.vars)
        self.vars[self.last_eval_id] = eval(expr, env, {})
        return self.vars[self.last_eval_id]

    def contains_comment(self, params):
        line, comment = params
        line = line.strip()
        if line != '':
            return self.parse_line(line)

    def parse_again(self, params):
        line = params[0]
        return self.parse_line(line)

    def eval_and_display(self, params):
        expr = params[0]
        self.display(self.evaluate(expr))

    def goto(self, params):
        # Since the current line number will be incremented right after
        # returning, we anticipate:
        label_id = int(params[0])
        self.current_line = self.labels[label_id]

    def conditional_jump(self, params):
        cond, expr = params
        if self.evaluate(cond):
            self.parse_line(expr)

    def if_then_else(self, params):
        # First, evaluate the condition.
        cond = params[0]
        if not self.evaluate(cond):
            # Jump to the relevant Else and ask to 'play' it.
            relevant_else_line = self.elses_for_ifs[self.current_line]
            # Anticipate the incrementation.
            self.current_line = relevant_else_line - 1
            self.play_this_else = relevant_else_line
        else:
            # Nothing special.
            pass

    def else_action(self, params):
        # Did we get here directly from an if?
        if self.play_this_else and self.play_this_else == self.current_line:
            # Yes.
            return self.parse_line(params[0])
        else:
            # No. Jump to the relevant IfEnd.
            # Anticipate the incrementation.
            try:
                self.current_line = self.ifends_for_elses[self.current_line] - 1
            except KeyError as e:
                self.display("Maybe you miss an %s for an %s." % (self.ifend_str, self.else_str))
                self.terminate()

    def for_step_action(self, params):
        """Implement 'for' loops.

        On the stack: 'for' line, control var name, step, end value.

        """
        if not self.loopers_stacks['for'] or \
           self.loopers_stacks['for'][-1][0] != self.current_line:
            # First time in this loop. Initialisation.
            var, init_val, end_val, step = params
            init_val = self.evaluate(init_val)
            end_val = self.evaluate(end_val)
            step = string_as_int_or_float(step)
            self.affectation((init_val, var))
            # Then push on the 'for' stack.
            self.loopers_stacks['for'].append((self.current_line, var, step, end_val))
        else:
            _, var, step, end_val = self.loopers_stacks['for'][-1]
            self.vars[var] = self.vars[var] + step
        # Increment the control var and check it against end_val.
        if self.vars[var] + step > end_val:
            # A 'next' knows when to not go back to the 'for' (continue)
            # when its 'for' is not on the for_stack.
            self.loopers_stacks['for'].pop()

    def next_action(self, params):
        for_line_gen = (k for k, v in self.loopers_pointers['for'].items() \
                                   if v == self.current_line)
        relevant_for_line = next(for_line_gen)
        # Is our 'for' still on the 'for' stack?
        if self.loopers_stacks['for'] and \
           self.loopers_stacks['for'][-1][0] == relevant_for_line:
            # Anticipate the incrementation.
            self.current_line = relevant_for_line - 1

    def while_action(self, params):
        """Implement 'while' loops.

        On the stack: 'whileend' line.

        """
        expr = params[0]
        whileend_line = self.loopers_pointers['while'][self.current_line]
        if not self.loopers_stacks['while'] or \
           self.loopers_stacks['while'][-1] != whileend_line:
            # First time in this loop. Push to the 'whileend' stack.
            self.loopers_stacks['while'].append(whileend_line)
        if not self.evaluate(expr):
            # A 'whileend' knows when to not go back to the 'while' (continue)
            # when it is not on the whileend_stack.
            self.current_line = self.loopers_stacks['while'].pop()

    def whileend_action(self, params):
        if self.loopers_stacks['while'][-1] == self.current_line:
            while_line_gen = (k for k, v in self.loopers_pointers['while'].items() \
                                       if v == self.current_line)
            relevant_while_line = next(while_line_gen)
            # Anticipate the incrementation.
            self.current_line = relevant_while_line - 1

class Casio(Calto):
    """Implementation of the Casio Basic Language.

    Differences with the 'official' syntax:
        -> is the single character arrow,
        ,  is the black triangle (not a separator, please add : if not at EOL),
        math expressions are evaluated as Python ones:
            == is used instead of =,

    """
    def __init__(self, screen = None):
        Calto.__init__(self, screen)
        self.split_chars = self.split_chars + ':'
        self.last_eval_id = 'Ans'
        self.label_pattern = '^Lbl (.*)$'
        self.parser_list = [
             ['^(.*)#(.*)', self.contains_comment],
             ['^(.*)\=\>(.*)$', self.conditional_jump],
             ['^If (.*)$', self.if_then_else],
             ['^Then (.*)$', self.parse_again],
             ['^Else (.*)$', self.else_action],
             ['^IfEnd$', lambda x: None],
             ['^For (.*)->(.*) To (.*) Step (.*)$', self.for_step_arg_sorter],
             ['^For (.*)->(.*) To (.*)$', self.for_step_arg_sorter],
             ['^Next$', self.next_action],
             ['^While (.*)', self.while_action],
             ['^WhileEnd', self.whileend_action],
             ['^"(.*)"$', lambda t: self.display(t[0])],
             ['^"(.*)"\?->(.*)$', self.prompt],
             ['^(.*)->(.*)$', self.affectation],
             ['^(.*),$', self.eval_and_display],
             ['^Lbl ', lambda x: None],
             ['^Goto (.*)$', self.goto],
             ]
        self.if_str = 'If '  # Space mandatory here because of IfEnd's.
        self.else_str = 'Else'
        self.ifend_str = 'IfEnd'
        # Space mandatory for 'While' because of WhileEnd's.
        self.loopers = [('for', 'For', 'Next'), ('while', 'While ', 'WhileEnd')]
        self.reset_loopers()
        self.compile_patterns()

    def for_step_arg_sorter(self, params):
        if len(params) == 4:
            step = params[3]
        else:
            step = '1'
        init_val, var, end_val = params[:3]
        return self.for_step_action((var, init_val, end_val, step))

    def parse_structures(self):
        # LOOP STRUCTURES
        # This means buildind a dict for each looper (see self.loopers) which
        # shows where to go when skipping the inner instructions.
        # These dicts are two way dicts:
        #     'for' lines can find their corresponding 'forend' line,
        #     'forend' lines can find their parent 'for' line.
        # The looper_stack contains (type, line_num) tuples. eg ('for', 11).
        looper_stack = []
        # Quick dict to find relevant end commands.
        end_cmds = dict()
        for looper_triple in self.loopers:
            end_cmds[looper_triple[0]] = looper_triple[2]
        # Now parse the lines.
        for line_num, line in enumerate(self.lines):
            line = line.lstrip()
            for looper, start_str, _ in self.loopers:
                if line.startswith(start_str):
                    looper_stack.append((looper, line_num))
                    break
            if looper_stack and line.startswith(end_cmds[looper_stack[-1][0]]):
                looper, parent_line = looper_stack.pop()
                self.loopers_pointers[looper][parent_line] = line_num
        # IF STRUCTURES
        # Parse if then else structures.
        # This means building two dicts:
        # self.elses_for_ifs show where to go if conds are False,
        #  (this one is used too if the Else branch was omitted)
        # self.ifends_for_elses show how to jump over Else blocs.
        if_stack = []
        for line_num, line in enumerate(self.lines):
            line = line.lstrip()
            if line.startswith(self.if_str):
                if_stack.append(line_num)
                continue
            if line.startswith(self.else_str):
                self.elses_for_ifs[if_stack[-1]] = line_num
                continue
            if line.startswith(self.ifend_str):
                relevant_if = if_stack.pop()
                if relevant_if in self.elses_for_ifs:
                    relevant_else = self.elses_for_ifs[relevant_if]
                    self.ifends_for_elses[relevant_else] = line_num
                else:
                    # The Else branch has been omitted, the Else
                    # line should be the IfEnd line.
                    self.elses_for_ifs[relevant_if] = line_num

    def affectation(self, params):
        val, varname = params
        if val == '?':
            val = self.ask("Please enter a value:")
        val_to_store = self.evaluate(str(val))
        self.vars[varname] = val_to_store
        return val_to_store


class TI(Calto):
    """Implementation of the Texas Instrument Basic Language.

    Differences with the 'official' syntax:

    """
    def __init__(self, screen = None):
        Calto.__init__(self, screen)
        self.split_chars = self.split_chars + ':'
        self.last_eval_id = 'Ans'
        self.label_pattern = '^Lbl (.*)$'
        self.parser_list = [
             ['^(.*)#(.*)', self.contains_comment],
             ['^If (.*)$', self.if_then_else],
             ['^Then (.*)$', self.parse_again],
             ['^Else (.*)$', self.else_action],
             ['^For\((.*),(.*),(.*),(.*)\)$', self.for_step_arg_sorter],
             ['^For\((.*),(.*),(.*)\)$', self.for_step_arg_sorter],
             ['^While (.*)', self.while_action],
             ['^End$', self.end_action],
             ['^Input ([^ ]*) (.*)$', self.affectation],
             ['^Input ([^ ]*)$', self.affectation],
             ['^Disp (.*)$', self.eval_and_display],
             ]
        self.if_str = 'If'
        self.else_str = 'Else'
        self.ifend_str = 'End'
        self.loopers = [('for', 'For', 'End'),
                        ('while', 'While', 'End'),
                        ('repeat','Repeat','End')]
        self.reset_loopers()
        self.compile_patterns()

    def for_step_arg_sorter(self, params):
        if len(params) == 4:
            step = params[3]
        else:
            step = '1'
        var, init_val, end_val = params[:3]
        return self.for_step_action((var, init_val, end_val, step))

    def end_action(self, params):
        back_lines = self.lines[:self.current_line]
        back_lines.reverse()
        for line in back_lines:
            if line.startswith('For'):
                return self.next_action(None)
                break
            elif line.startswith('While'):
                return self.whileend_action(None)
                break
            elif line.startswith('If'):
                # Nothing to do for an End of an If, except returning.
                return

    def parse_structures(self):
        # LOOP and IF structures have to be parsed in one go 'thanks'
        # to them having the same ending word: 'End'.
        # Loops:
        # This means buildind a dict for each looper (see self.loopers) which
        # shows where to go when skipping the inner instructions.
        # These dicts are two way dicts:
        #     'for' lines can find their corresponding 'forend' line,
        #     'forend' lines can find their parent 'for' line.
        # The looper_stack contains (type, line_num) tuples. eg ('for', 11).
        # If-then-else:
        # This means building two dicts:
        # self.elses_for_ifs show where to go if conds are False,
        #  (this one is used too if the Else branch was omitted)
        # self.ifends_for_elses show how to jump over Else blocs.
        struct_stack = []
        # Now parse the lines.
        for line_num, line in enumerate(self.lines):
            line = line.lstrip()
            # Loops:
            for looper, start_str, _ in self.loopers:
                if line.startswith(start_str):
                    struct_stack.append((looper, line_num))
                    break
            # If-then-else:
            if line.startswith(self.if_str):
                struct_stack.append(('if', line_num))
                continue
            if line.startswith(self.else_str):
                self.elses_for_ifs[struct_stack[-1][1]] = line_num
                continue
            # Ends:
            if struct_stack and line.startswith('End'):
                struct, parent_line = struct_stack.pop()
                if struct != 'if':
                    # Loops:
                    self.loopers_pointers[struct][parent_line] = line_num
                else:
                    # If-then-else:
                    relevant_if = parent_line
                    if relevant_if in self.elses_for_ifs:
                        relevant_else = self.elses_for_ifs[relevant_if]
                        self.ifends_for_elses[relevant_else] = line_num
                    else:
                        # The Else branch has been omitted, the Else
                        # line should be the IfEnd line.
                        self.elses_for_ifs[relevant_if] = line_num

    def affectation(self, params):
        if len(params) == 2:
            val, varname = params
        else:
            val = self.ask("Please enter a value:")
            varname = params[0]
        val_to_store = self.evaluate(str(val))
        self.vars[varname] = val_to_store
        return val_to_store


def tests(calto_types=None):
    FAIL_STR = "*" * 40 + "\n%s should give %s\nbut give %s"
    tests_suite = {
        "Casio": [
            ('', []),
            (':', []),
            ('# this is a comment', []),
            ('2# this is a comment', ['2']),
            ('2  # this is a comment', ['2']),
            ('"test"', ['test']),
            ('2+2', ['4']),
            ('2->X:X*2.54', ['5.08']),
            ('2->X  # comment:X*2.54', ['5.08']),
            ('2->A:A+5->A', ['7']),
            ('2->A:A-5->A', ['-3']),
            ('2.1->A:A+5->A', ['7.1']),
            ('2->A:A**2:Ans**2', ['16']),
            ('1.1->A:A**2:Ans**2', ['1.4641']),
            ('2->A:A**2,:Ans**2', ['4', '16']),
            ('1.1->A:A**2,:Ans**2', ['1.21', '1.4641']),
            ('2->A:Goto 1:3->A:Lbl 1:A', ['2']),
            ('2->A:Lbl 1:A', ['2']),
            ('1=>1,:0', ['1', '0']),
            ('0=>1,:0', ['0']),
            ('2->A:If A>0:Then 10*A,:Else 9*A,:IfEnd:Ans*2', ['20','40']),
            ('2->A:If A>0:Then 10*A,:10*A,:Else 9*A,:IfEnd:Ans*2', ['20','20','40']),
            ('2->A:If A<0:Then 10*A,:Else 9*A,:IfEnd:Ans*2', ['18','36']),
            ('2->A:If A<0:Then 10*A,:Else 9*A,:IfEnd', ['18']),
            ('2->A:If A>0:Then 2*A->A:IfEnd:A', ['4']),
            ('2->A:If A<0:Then 2*A->A:IfEnd:A', ['2']),
            ('2->A:If A>0:Then 10*A,:If 0==0:Then 10,:Else 20,:Else 9*A,:IfEnd:Ans*2', ['20','10',"Maybe you miss an IfEnd for an Else."]),
            ('2->A:If A>0:Then 10*A,:If 0==0:Then 10,:Else 20,:IfEnd:Else 9*A,:IfEnd:Ans*2', ['20','10','20']),
            ('2->A:If A>0:Then 10*A,:If 0!=0:Then 10,:Else 20,:IfEnd:Else 9*A,:IfEnd:Ans*2', ['20','20','40']),
            ('2->A:If A<0:Then 10*A,:Else 9*A,:If 0==0:Then 10,:Else 20,:IfEnd:IfEnd:Ans*2', ['18','10','20']),
            ('2->A:If A<0:Then 10*A,:Else 9*A,:If 0!=0:Then 10,:Else 20,:IfEnd:IfEnd:Ans*2', ['18','20','40']),
            ('For 1->A To 5:A**2->B:B,:Next', ['1', '4', '9', '16', '25']),
            ('For 1->A To 10 Step 2:A**2->B:B,:Next', ['1', '9', '25', '49', '81']),
            ('For 1->A To 1:A**2->B:B,:Next', ['1']),
            ('For 1->A To 1 Step 2:A**2->B:B,:Next', ['1']),
            ('For 1->A To 3:For 1->B To 2:A,:B,:Next:Next', ['1','1', '1','2', '2','1', '2','2', '3','1', '3','2']),
            ('1->A:While A<1:A**2,:A+1->A:WhileEnd', []),
            ('1->A:While A<2:A**2,:A+1->A:WhileEnd', ['1']),
            ('1->A:While A<5:A**2,:A+1->A:WhileEnd', ['1', '4', '9', '16']),
            ('1->A:While A<5:A**2,:A+1->A:WhileEnd:A/2.', ['1', '4', '9', '16', '2.5']),
            ('1->A:1->B:While A<4:While B<3:A,:B,:B+1->B:WhileEnd:1->B:A+1->A:WhileEnd', ['1','1', '1','2', '2','1', '2','2', '3','1', '3','2']),
            ],
        "TI": [
            ('', []),
            (':', []),
            ('# this is a comment', []),
            ('2# this is a comment', ['2']),
            ('2  # this is a comment', ['2']),
            ('Input 2 X:X*2.54', ['5.08']),
            ('Input 2 A:Input A+5 A', ['7']),
            ('Input 2.1 A:Input A+5 A', ['7.1']),
            ('For(A,1,5):Input A**2 B:Disp B:End', ['1', '4', '9', '16', '25']),
            ('For(A,1,10,2):Input A**2 B:Disp B:End', ['1', '9', '25', '49', '81']),
            ('For(A,1,1):Input A**2 B:Disp B:End', ['1']),
            ('For(A,1,1,2):Input A**2 B:Disp B:End', ['1']),
            ('For(A,1,3):For(B,1,2):Disp A:Disp B:End:End', ['1','1', '1','2', '2','1', '2','2', '3','1', '3','2']),
            ('Input 1 A:While A<1:Disp A**2:Input A+1 A:End', []),
            ('Input 1 A:While A<2:Disp A**2:Input A+1 A:End', ['1']),
            ('Input 1 A:While A<5:Disp A**2:Input A+1 A:End', ['1', '4', '9', '16']),
            ('Input 1 A:While A<5:Disp A**2:Input A+1 A:End:A/2.', ['1', '4', '9', '16', '2.5']),
            ('Input 1 A:Input 1 B:While A<4:While B<3:Disp A:Disp B:Input B+1 B:End:Input 1 B:Input A+1 A:End', ['1','1', '1','2', '2','1', '2','2', '3','1', '3','2']),
            ('Input 2 A:If A>0:Disp 10*A:Else Disp 9*A:End:Ans*2', ['20','40']),
            ('Input 2 A:If A<0:Disp 10*A:Else Disp 9*A:End:Ans*2', ['18','36']),
            ('Input 2 A:If A<0:Disp 10*A:Else Disp 9*A:End', ['18']),
            ('Input 2 A:If A>0:Input 2*A A:End:A', ['4']),
            ('Input 2 A:If A<0:Input 2*A A:End:A', ['2']),
            ('Input 2 A:If A>0:Then Disp 10*A:Else Disp 9*A:End:Ans*2', ['20','40']),
            ('Input 2 A:If A<0:Then Disp 10*A:Else Disp 9*A:End:Ans*2', ['18','36']),
            ('Input 2 A:If A<0:Then Disp 10*A:Else Disp 9*A:End', ['18']),
            ('Input 2 A:If A>0:Then Input 2*A A:End:A', ['4']),
            ('Input 2 A:If A<0:Then Input 2*A A:End:A', ['2']),
            ('Input 2 A:If A>0:Then Disp 10*A:If 0==0:Then Disp 10:Else Disp 20:Else Disp 9*A:End:Ans*2', ['20','10',"Maybe you miss an End for an Else."]),
            ('Input 2 A:If A>0:Then Disp 10*A:If 0==0:Then Disp 10:Else Disp 20:End:Else Disp 9*A:End:Ans*2', ['20','10','20']),
            ('Input 2 A:If A>0:Then Disp 10*A:If 0!=0:Then Disp 10:Else Disp 20:End:Else Disp 9*A:End:Ans*2', ['20','20','40']),
            ('Input 2 A:If A<0:Then Disp 10*A:Else Disp 9*A:If 0==0:Then Disp 10:Else Disp 20:End:End:Ans*2', ['18','10','20']),
            ('Input 2 A:If A<0:Then Disp 10*A:Else Disp 9*A:If 0!=0:Then Disp 10:Else Disp 20:End:End:Ans*2', ['18','20','40']),
            ('For(A,1,1):If A==0:Then Input A**2 A:End:Disp A:End', ['1']),
            ],
        }
    if not calto_types:
        calto_types = tests_suite.keys()
    for calto_type in calto_types:
        calto_class = eval(calto_type)
        mycalto = calto_class()
        for source, expected_results in tests_suite[calto_type]:
            mycalto.reset()
            mycalto.testing = True
            mycalto.run(source)
            if expected_results != mycalto.displays:
                print(FAIL_STR % (source, str(expected_results), str(mycalto.displays)))


if __name__ == '__main__':
    if len(sys.argv) == 1:
        # No arg on the command line.
        tests()
    elif len(sys.argv) == 2:
        # One arg: a file to run.
        filename = sys.argv[1]
        with open(filename, 'r') as f:
            source = f.read().strip()
        # File extension defines the calculator.
        ext2calto = {
                '.casio': Casio,
                '.ti': TI,
                }
        _, ext = os.path.splitext(filename)
        mycalto = ext2calto[ext]()
        mycalto.run(source)
    elif sys.argv[1] in ['-t', '--test']:
        # Run the tests for a specific calculator.
        tests([sys.argv[2]])
    else:
        print("Usage:")
        print("Run a file:        " + sys.argv[0] + " [file]")
        print("Run all the tests: " + sys.argv[0])
        print("Run a test suite:  " + sys.argv[0] + " -t calc_type")
        print(" Where calc_type is either Casio or TI.")
